from signum.core.base_container import SignalContainer
from signum.core.containers import TimeDomainSignal, FreqDomainSignal
from signum.tools.scale_manager import ScaleManager


__version__ = '0.3.0'



import numpy as np
import logging

from signum.tools.scale_manager import ScaleManager
from signum.core.base_container import SignalContainer, SCA


logger = logging.getLogger(__name__)


class BaseTimeDomainSignal(SignalContainer):
    ATTRIBUTES = SignalContainer.ATTRIBUTES + [SCA('_t_sampling', 1, True)]

    def __new__(cls, *args, f_sampling: float = 1, f_sampling_unit='Hz', **kwargs):
        kwargs['x_unit'] = kwargs.pop('x_unit', 's')
        kwargs['x_description'] = kwargs.pop('x_description', 'Time')
        obj = super().__new__(cls, *args, **kwargs)

        # set sampling time
        f_sampling, f_sampling_unit = ScaleManager.rescale_to_basic(f_sampling, f_sampling_unit)
        obj.t_sampling = 1./f_sampling  # using the t sampling setter

        return obj

    @property
    def resolution(self):
        return self._base_resolution * self._t_sampling

    @resolution.setter
    def resolution(self, val):
        self._resolution_setter_(val)

    def _resolution_setter_(self, val):
        self._t_sampling = val
        self._resolution_defaults_setter()

    @property
    def t_sampling(self):
        return self.resolution

    @t_sampling.setter
    def t_sampling(self, val):
        self._resolution_setter_(val)

    @property
    def f_sampling(self):
        return 1./self.t_sampling

    @f_sampling.setter
    def f_sampling(self, val):
        self._resolution_setter_(1./val)

    @staticmethod
    def check_metadata(arr1, arr2, require_type=False):
        for i, arr in enumerate([arr1, arr2]):
            if not isinstance(arr, BaseTimeDomainSignal):
                m = f'Input array #{i+1} is not an instance of BaseTimeDomainSignal'
                if require_type:
                    raise TypeError(m)
                logger.debug(m)
                return

        if not np.isclose(arr1.f_sampling, arr2.f_sampling):
            raise ValueError(f"Inputs sampling frequencies mismatch: {arr1.f_sampling:.2E} and {arr2.f_sampling:.2E}")



import matplotlib.pyplot as plt
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from signum import SignalContainer

from signum.plotting.plotter import Plotter


class IQPlotter(Plotter):
    def __init__(self, figsize=(8, 6), sharey='all', unit='', x_unit='', **kwargs):
        super().__init__(n_rows=2, n_cols=1, sharex='all', sharey=sharey, figsize=figsize, **kwargs)

        self.i_ax.set_ylabel("Real")
        self.q_ax.set_ylabel("Imag")

        self.i_ax.set_y_unit(unit)
        self.q_ax.set_y_unit(unit)
        self.q_ax.set_x_unit(x_unit)

    @property
    def i_ax(self) -> plt.Axes:
        return self.axes[0, 0]

    @property
    def q_ax(self) -> plt.Axes:
        return self.axes[1, 0]

    def _add_line(self, signal: 'SignalContainer', **kwargs):
        i_lines = self.i_ax.plot(signal.real.T, **kwargs)
        q_lines = self.q_ax.plot(signal.imag.T, **kwargs)

        return i_lines, q_lines




import logging

import numpy as np
from matplotlib import mlab

from signum.core.freq_domain_signal import BaseFreqDomainSignal as BaseFreqDomainSignal
from signum.core.time_domain_signal import BaseTimeDomainSignal as BaseTimeDomainSignal
from signum.tools.scale_manager import ScaleManager

logger = logging.getLogger(__name__)


class TimeDomainSignal(BaseTimeDomainSignal):
    def csd(self, other=None, **kwargs):
        """Cross spectral density"""

        if other is None:
            description = f"{self.description}: power spectral density"
            meta = {**self.meta, 'original_description': self.description}
            other = self

        else:
            # check that the operation is applicable to the inputs
            self.check_metadata(self, other, require_type=True)

            description = f"{self.description} & {other.description}: cross spectral density"
            meta = {**self.meta, **other.meta, 'original_descriptions': [self.description, other.description]}

        # calculate the spectrum
        spectrum, frequencies = mlab.csd(other, self, **kwargs, Fs=self.f_sampling)

        spectrum = FreqDomainSignal(spectrum, f_resolution=frequencies[1] - frequencies[0], x_start=frequencies[0],
                                    description=description, meta=meta, unit=self.unit)

        return spectrum

    def psd(self, **kwargs):
        """Power spectral density"""

        return self.csd(other=None, **kwargs)

    def fft(self, axis=-1, description='', **kwargs):
        self._check_x_axis_spacing("a Fourier transform")

        f = np.fft.fft(self, axis=axis, **kwargs)

        ns = self.shape[axis]
        f_resolution = self.f_sampling/ns
        description = description or (f"FFT of {self.description}" if self.description else '')

        f = FreqDomainSignal(f, f_resolution=f_resolution, description=description, meta=self.meta, unit=self.unit)

        return f


class FreqDomainSignal(BaseFreqDomainSignal):
    def ifft(self, axis=-1, description='', **kwargs):
        self._check_x_axis_spacing("an inverse Fourier transform")

        ns = self.shape[axis]
        f_sampling = self.f_resolution * ns
        description = description or (f"Inverse FFT of {self.description}" if self.description else '')

        if self.x_axis[0]:
            logger.warning(f"The first element of the array appears not to be the zero-frequency term (based on the "
                           f"frequency axis: {ScaleManager.display(self.x_axis[0], 'Hz')})")

        t = np.fft.ifft(self, axis=axis, **kwargs)
        t = TimeDomainSignal(t, meta=self.meta, unit=self.unit, description=description, f_sampling=f_sampling)
        return t

    def fftshift(self):
        self._check_x_axis_spacing("FFT shift")

        f = FreqDomainSignal(np.fft.fftshift(self))
        f.copy_attributes(self)
        f.x_start = self.x_start - self.x_axis[self.shape[-1] // 2]

        return f

    def ifftshift(self):
        self._check_x_axis_spacing("Inverse FFT shift")

        f = FreqDomainSignal(np.fft.ifftshift(self))
        f.copy_attributes(self)
        f.x_start = self.x_axis[self.shape[-1] // 2]

        return f


if __name__ == '__main__':
    # time domain
    tdata1 = TimeDomainSignal(np.random.rand(12), f_sampling=40e6, description='Signal 1')
    tdata2 = TimeDomainSignal(np.arange(12).reshape(1, -1), f_sampling=1, unit='V', description='Signal 2')
    tdata3 = TimeDomainSignal(np.arange(12), f_sampling=40e6, description='Signal 3')

    tdata1.display()

    psd = tdata1.psd()
    psd.display()

    csd = tdata1.csd(tdata3)
    csd.display()

    # freq domain
    fdata1 = FreqDomainSignal(np.random.rand(12) + 1j * np.random.rand(12), f_resolution=10, unit='mV',
                              description="Frequency response")
    fdata2 = FreqDomainSignal(np.arange(12).reshape(1, -1), unit='V', description="H")

    fdata2.reshape(-1).display()
    fdata1.display()
    fdata1.display(plot_type='bode', db_scale=True)
    fdata1.display(plot_type='nyquist')



from signum.core.base_container import SignalContainer, SCA


class BaseFreqDomainSignal(SignalContainer):
    ATTRIBUTES = SignalContainer.ATTRIBUTES + [SCA('_f_resolution', 1, True)]

    # alias for the property
    f_resolution = SignalContainer.resolution

    def __new__(cls, *args, f_resolution: float=None, **kwargs):
        obj = super().__new__(cls, *args, **kwargs, x_unit='Hz', x_description='Frequency')

        if f_resolution is not None:
            obj.f_resolution = f_resolution  # using the property setter

        return obj

    @property
    def resolution(self):
        return self._base_resolution * self._f_resolution

    @resolution.setter
    def resolution(self, val):
        self._resolution_setter_(val)

    def _resolution_setter_(self, val):
        self._f_resolution = val
        self._resolution_defaults_setter()

    @property
    def f_sampling(self):
        return self.resolution

    @f_sampling.setter
    def f_sampling(self, val):
        self._resolution_setter_(val)





import matplotlib.pyplot as plt
from matplotlib.projections.polar import PolarAxes
import logging

from signum.plotting.scaled_formatter import ScaledFormatter
from signum.tools.scale_manager import ScaleManager


_NOT_GIVEN_ = object()

logger = logging.getLogger(__name__)


class ScaledAxes:
    def __init__(self, ax: plt.Axes, x_unit='', y_unit='', x_label=None, y_label=None):
        if not isinstance(ax, plt.Axes):
            raise TypeError(f"Expected matplotlib.pyplot.Axes instance, got {type(ax)=}")

        self._ax = ax
        if not isinstance(ax, PolarAxes):
            ax.xaxis.set_major_formatter(ScaledFormatter(x_unit))
        ax.yaxis.set_major_formatter(ScaledFormatter(y_unit))

        if x_label:
            self.set_xlabel(x_label)
        if y_label:
            self.xet_ylabel(y_label)

    def __getattr__(self, item):
        return getattr(self._ax, item)

    @property
    def ax(self):
        return self._ax

    @property
    def x_ax_formatter(self):
        return self.ax.xaxis.get_major_formatter()

    @property
    def y_ax_formatter(self):
        return self.ax.yaxis.get_major_formatter()

    def get_x_unit(self):
        return self.x_ax_formatter.base_unit

    def set_x_unit(self, x_unit):
        self.x_ax_formatter.base_unit = x_unit

    def get_y_unit(self):
        return self.y_ax_formatter.base_unit

    def set_y_unit(self, y_unit):
        self.y_ax_formatter.base_unit = y_unit

    def get_units(self):
        return self.get_x_unit(), self.get_y_unit()

    def set_units(self, x_unit, y_unit):
        self.set_x_unit(x_unit)
        self.set_y_unit(y_unit)

    def plot(self, *args, add_legend=True, **kwargs):
        if len(args) == 3:
            x_data, y_data, fmt = args
        elif len(args) == 2:
            if isinstance(args[1], str):
                y_data, fmt = args
                x_data = None
            else:
                x_data, y_data = args
                fmt = None
        elif len(args) == 1:
            y_data, = args
            x_data = None
            fmt = None
        else:
            raise TypeError(f"Expected 1 to 3 positional arguments, got {len(args)}")

        try:
            x_unit = x_data.unit if x_data is not None else y_data.x_unit
            y_unit = y_data.unit
            x_data = x_data if x_data is not None else y_data.x_axis

        except AttributeError:
            raise TypeError(f"Expected a SignalContainer instance, got {type(y_data)}")

        x_data, y_data = self._fix_units(x_data, y_data, x_unit, y_unit)

        plot_args = (x_data, y_data, fmt) if fmt is not None else (x_data, y_data)
        lines = self.ax.plot(*plot_args, **kwargs)

        if add_legend and 'label' in kwargs:
            self.add_legend()

        return lines

    def add_legend(self, fancybox=True, framealpha=0.5, **kwargs):
        self.ax.legend(fancybox=fancybox, framealpha=framealpha, **kwargs)

    def _fix_units(self, data_x, data_y, data_x_unit, data_y_unit):
        data = {'x': data_x, 'y': data_y}
        data_units = {'x': data_x_unit, 'y': data_y_unit}

        for xy in 'xy':
            if isinstance(getattr(self.ax, f'{xy}axis').get_major_formatter(), ScaledFormatter):
                ax_unit = getattr(self, f'get_{xy}_unit')()
                data_unit = data_units[xy]
                if ax_unit:
                    if ax_unit != data_unit:
                        _, ax_unit_core = ScaleManager.split_unit(ax_unit)
                        _, data_unit_core = ScaleManager.split_unit(data_unit)
                        if ax_unit_core != data_unit_core:
                            raise ValueError(f"{xy} units of the axes and data do not agree: {ax_unit=}, {data_unit=}")
                        order_change = ScaleManager.order_from_unit(ax_unit) - ScaleManager.order_from_unit(data_unit)
                        data[xy], _ = ScaleManager.rescale(data[xy], data_unit, order_change)
                else:
                    logger.debug(f"Setting {xy} unit of the axes to data {xy} unit: {data_unit}")
                    getattr(self, f'set_{xy}_unit')(data_unit)

        return data['x'], data['y']



from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np

if TYPE_CHECKING:
    from signum import SignalContainer

from signum.plotting.scaled_axes import ScaledAxes


class Plotter:
    def __init__(self, n_rows=1, n_cols=1, title=None, **kwargs):
        fig, axes = plt.subplots(n_rows, n_cols, **kwargs)
        if n_rows == 1 and n_cols == 1:
            axes = np.array([axes])
        axes = axes.reshape(n_rows, n_cols)

        for row in range(n_rows):
            for col in range(n_cols):
                axes[row, col] = ScaledAxes(axes[row, col])

        self.fig = fig
        self.axes = axes
        self.lines = {}
        self.plot_data = {}

        if title:
            fig.suptitle(title)

        for ax in axes.flatten():
            ax.grid(color='lightgrey')

    def add_legend(self, **kwargs):
        """Add legend to each of the main grid axes."""

        for ax in self.axes.flatten():
            ax.add_legend(**kwargs)

    def show_fig(self, **kwargs):
        self.fig.show(**kwargs)

    @staticmethod
    def show_all(**kwargs):
        plt.show(**kwargs)

    def add_line(self, signal: 'SignalContainer', add_legend=True, **kwargs):
        if signal.description and 'label' not in kwargs:
            kwargs['label'] = signal.description

        lines = self._add_line(signal, **kwargs)

        if add_legend and 'label' in kwargs:
            self.add_legend()

        k = kwargs.get('label', f'line {len(self.lines)}')
        self.lines[k] = lines
        self.plot_data[k] = signal

    def _add_line(self, signal, **kwargs):
        raise NotImplementedError


class SimplePlotter(Plotter):
    def __init__(self, x_label='', y_label='', **kwargs):
        super().__init__(n_rows=1, n_cols=1, **kwargs)

        self.ax.set_xlabel(x_label)
        self.ax.set_ylabel(y_label)

    @property
    def ax(self):
        return self.axes[0, 0]

    def _add_line(self, signal: 'SignalContainer', **kwargs):
        if np.iscomplexobj(signal):
            raise ValueError("Can't plot complex data on a SimplePlotter")

        lines = self.ax.plot(signal, **kwargs)

        return lines



import numpy as np
import matplotlib.pyplot as plt
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from signum import SignalContainer

from signum.plotting.plotter import Plotter


class NyquistPlotter(Plotter):
    def __init__(self, figsize=(6, 6), unit='', **kwargs):
        super().__init__(n_rows=1, n_cols=1, figsize=figsize, **kwargs)

        self.ax.set_xlabel("I component")
        self.ax.set_ylabel("Q component")
        self.ax.set_x_unit(unit)
        self.ax.set_y_unit(unit)

        self.ax.set_aspect("equal")
        self.ax.axvline(0, color='k', zorder=1)
        self.ax.axhline(0, color='k', zorder=1)

    @property
    def ax(self) -> plt.Axes:
        return self.axes[0, 0]

    def _add_line(self, signal: 'SignalContainer', set_equal_limits=True, **kwargs):
        lines = self.ax.plot(signal.real.T, signal.imag.T, **kwargs)

        self.ax.relim()

        if set_equal_limits:
            lim = np.max(np.abs([*self.ax.get_xlim(), *self.ax.get_ylim()]))
            self.ax.set_xlim(-lim, lim)
            self.ax.set_ylim(-lim, lim)

        return lines



import matplotlib.pyplot as plt
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from signum import SignalContainer

from signum.plotting.plotter import Plotter


class PolarPlotter(Plotter):
    def __init__(self, figsize=(6, 6), db_scale=False, unit='', **kwargs):
        super().__init__(n_rows=1, n_cols=1, figsize=figsize, subplot_kw={'projection': 'polar'}, **kwargs)

        self._db_scale = db_scale

        self.ax.set_y_unit(unit)

        self.ax.grid(color='lightgrey')

    @property
    def ax(self) -> plt.Axes:
        return self.axes[0, 0]

    def _add_line(self, signal: 'SignalContainer', set_equal_limits=True, **kwargs):
        mag = signal.magnitude_db if self._db_scale else signal.magnitude
        phase = signal.get_phase(rad=True)

        lines = self.ax.plot(phase.T, mag.T, **kwargs)

        return lines



import numpy as np

from signum.plotting.bode_plotter import BodePlotter
from signum.plotting.iq_plotter import IQPlotter
from signum.plotting.nyquist_plotter import NyquistPlotter
from signum.plotting.plotter import Plotter, SimplePlotter
from signum.plotting.polar_plotter import PolarPlotter

complex_plotters = {
    'bode': BodePlotter,
    'iq': IQPlotter,
    'nyquist': NyquistPlotter,
    'polar': PolarPlotter
}

plotters = {
    'simple': SimplePlotter,
    **complex_plotters
}


def get_plotter_class(plotter_type, complex_only=False):
    plotter_dict = complex_plotters if complex_only else plotters

    if plotter_type in plotter_dict:
        return plotter_dict[plotter_type]

    plotter_names = list(plotter_dict.keys())
    raise ValueError(f"Unknown {'complex ' if complex_only else ''}plotter type: '{plotter_type}'; "
                     f"choose from: {plotter_names}")


def get_plotter(plotter_type, complex_only=False, **kwargs):
    return get_plotter_class(plotter_type, complex_only=complex_only)(**kwargs)


def plot_signal(signal, plot_type: str = None, show: bool = True, title: str = None, plot_kwargs=None, **kwargs):
    if signal.size < 2:
        kwargs['marker'] = kwargs.get('marker', 'o')  # put a single dot if there is only one data point

    is_complex = np.iscomplexobj(signal)
    if plot_type is None:
        plot_type = 'bode' if is_complex else 'simple'
    ptr = get_plotter(plot_type, complex_only=is_complex, title=title, **(plot_kwargs or {}))

    ptr.add_line(signal, **kwargs)

    if show:
        ptr.show_fig()

    return ptr



import matplotlib.pyplot as plt
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from signum import SignalContainer

from signum.plotting.plotter import Plotter


class BodePlotter(Plotter):
    def __init__(self, figsize=(8, 6), db_scale: bool = False, rad: bool = False, unwrapped: bool = False,
                 unit='', x_unit='', **kwargs):
        super().__init__(n_rows=2, n_cols=1, sharex='all', figsize=figsize, **kwargs)

        self.amplitude_ax.set_ylabel(f"Amplitude")
        self.amplitude_ax.set_y_unit('dB' if db_scale else unit)
        self.phase_ax.set_ylabel(f"Phase")
        self.phase_ax.set_y_unit('rad' if rad else 'deg')
        self.phase_ax.set_x_unit(x_unit)

        self._db_scale = db_scale
        self._rad = rad
        self._unwrapped = unwrapped

    @property
    def amplitude_ax(self) -> plt.Axes:
        return self.axes[0, 0]

    @property
    def phase_ax(self) -> plt.Axes:
        return self.axes[1, 0]

    def _add_line(self, signal: 'SignalContainer', **kwargs):
        mag = signal.magnitude_db if self._db_scale else signal.magnitude
        amplitude_lines = self.amplitude_ax.plot(mag.T, **kwargs)

        phase = signal.get_phase(rad=self._rad, unwrapped=self._unwrapped)
        phase_lines = self.phase_ax.plot(phase.T, **kwargs)

        return amplitude_lines, phase_lines




import numpy as np
from matplotlib.ticker import ScalarFormatter

from signum.tools.scale_manager import ScaleManager


class ScaledFormatter(ScalarFormatter):

    def __init__(self, base_unit, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.base_unit = base_unit

    def get_offset_order(self):
        return int(np.log10(self.offset))

    def get_offset(self):
        if len(self.locs) == 0:
            return ''

        if not self.base_unit:
            return super().get_offset()

        if self.offset:
            offset_str = ScaleManager.display(self.offset, unit=self.base_unit,
                                              order_change=0 if self.no_rescale else None)
            if self.offset > 0:
                offset_str = '+' + offset_str
        else:
            offset_str = ''

        if self.orderOfMagnitude:
            oom_val, oom_unit = ScaleManager.rescale(10 ** self.orderOfMagnitude, self.base_unit)
            oom_str = oom_unit if oom_val == 1 else ScaleManager.display(oom_val, oom_unit)
        else:
            oom_str = self.base_unit

        s = ' '.join((oom_str, offset_str))

        return self.fix_minus(s)

    @property
    def no_rescale(self):
        return self.base_unit.lower() in ['db', 'deg', 'rad', 'none', '']

    def _set_order_of_magnitude(self):
        if len(self.locs) == 0 or self.no_rescale:
            self.orderOfMagnitude = 0
        else:
            # Restrict to visible ticks (this line comes from the ScalarFormatter code)
            vmin, vmax = sorted(self.axis.get_view_interval())

            order = int(np.log10(max([abs(vmin - self.offset), abs(vmax - self.offset)])))
            self.orderOfMagnitude = 3 * divmod(order, 3)[0]

    def __call__(self, x, pos=None):
        if len(self.locs) == 0:
            return ''

        xp = (x - self.offset) / (10. ** self.orderOfMagnitude)
        if abs(xp) < 1e-8:
            xp = 0
        s = self.format % xp
        return self.fix_minus(s)
